/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : AirScheduleProblem.cc                                    */
/*                                                                       */
/*       Date : May     4, 1997                                          */
/* Updated at : March  14, 1999                                          */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/
/* Representation of the Flight Schedule 
*/
/*************************************************************************/

#include <iomanip.h>
#include <string.h>
#include <stdio.h>

#include <deffunc.h>

#include <myQueue.h>

#include "AirScheduleProblem.h"
#include "Flight.h"
#include "City.h"
#include "Column.h"
#include "FlightGraph.h"
#include "PathBuilder.h"


/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::AirScheduleProblem()
{  
   bIndex= cIndex=numberOfFlights = 0;
   costPerMinute = costPerOvernight = 0;
   loadBalanceFactor = 1.0;
   minimumTime = 20.0;
   cities = BaseCities = NULL;
   FlightList = NULL;
   myEdgeGraph=NULL;
   costPerMinuteUnderUtilization=0.2;
   difficulty = NULL;

}// end of constructor

/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::AirScheduleProblem(const AirScheduleProblem& RHS)
{  myFlights = RHS.myFlights;

   bIndex = RHS.bIndex;
   BaseCities = new City[bIndex];
   for(unsigned i=0;i<bIndex;i++) BaseCities[i] = RHS.BaseCities[i];

   cIndex = RHS.cIndex;
   cities = new City[cIndex];
   for(i=0;i<cIndex;i++) cities[i] = RHS.cities[i];

      costPerMinute = RHS.costPerMinute;
   costPerOvernight = RHS.costPerOvernight;
        minimumTime = RHS.minimumTime;
    numberOfFlights = RHS.numberOfFlights;
    costPerMinuteUnderUtilization = RHS.costPerMinuteUnderUtilization;

   FlightList = new Flight[numberOfFlights];
   for(i=0;i<numberOfFlights;i++) FlightList[i] = RHS.FlightList[i];
   difficulty = new float[numberOfFlights];
   for(i=0;i<numberOfFlights;i++) difficulty[i] = 1.0;
   myEdgeGraph=NULL;

}// end of COPY constructor


/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::~AirScheduleProblem()
{  cout << " de allocate AirScheduleProblem() " << endl; }

/*
//
// Sets the deadHeadingCostRatio.
// It can not be less then 1.0
//
*/

void AirScheduleProblem::setDeadheadingCostRatio(float _cost) 
{ deadheadingCostRatio = (_cost>=1.0?_cost:deadheadingCostRatio); }

/*
//
// Sets the per minute cost.
//
*/

void AirScheduleProblem::setCostPerMinute(float _cost)
{  costPerMinute = _cost; }

void AirScheduleProblem::setCostPerMinuteUnderUtilization(float _cost)
{  costPerMinuteUnderUtilization = _cost; }

/*
// Sets the load balancing factor.
*/

void AirScheduleProblem::setLoadBalancingFactor(float _cost)
{  loadBalanceFactor = _cost; }

/*
//
// Sets the overnight stay cost.
//
*/

void AirScheduleProblem::setCostPerOvernight(float _cost)
{  costPerOvernight = _cost; }

/*
//
// Sets the minimum amount of time we should put between consecutive
// flights.
*/

void AirScheduleProblem::setMinimumTime(float _cost)
{  minimumTime = _cost; }

/*
//
// Returns the deadHeadingCostRatio.
//
*/

float AirScheduleProblem::getDeadheadingCostRatio() const 
{ return deadheadingCostRatio;}

float AirScheduleProblem::getTotalFlyingTime() const 
{ 
   float ret = 0; 

   for(int i=0;i<numberOfFlights;i++) 
       ret += (FlightList[i].getArrivalDate()-FlightList[i].getDepartureDate());

   return ret;

}// end of getTotalFlyingTime()

AirScheduleProblem::getFlightIndexFromFlightNumber(unsigned FlightNumber)
{ 
   int ret = -1;
   int i=0, found=0;

   while(i<numberOfFlights && !found)
   {
      if( FlightNumber == FlightList[i].getFlightNumber() ) 
      { found = 1; ret=i; }
      else i++;

   }// end of while

   return ret;
}

/*************************************************************************/
// Returns the start date of the plan horizon
/*************************************************************************/
Date AirScheduleProblem::getStartOfPlanHorizon() const
{
   Date ret;
   if(numberOfFlights) ret = FlightList[0].getDepartureDate();
   return ret;
}// end of getStartOfPlanHorizon()

/*************************************************************************/
// Returns the end date of the plan horizon
/*************************************************************************/

Date AirScheduleProblem::getEndOfPlanHorizon() const
{
   Date ret;
   if(numberOfFlights) ret = FlightList[numberOfFlights-1].getArrivalDate();
   return ret;
}// end of getEndOfPlanHorizon()

/*************************************************************************/
// Returns the per minute cost for underutilization where crew stays
// without assignment.
/*************************************************************************/

float AirScheduleProblem::getCostPerMinuteUtilization() const 
{ return costPerMinuteUnderUtilization;}

/*************************************************************************/
// Returns the per minute cost.
/*************************************************************************/

float AirScheduleProblem::getCostPerMinute() const 
{ return costPerMinute;}

/*************************************************************************/
/** Denotes the amount of punishment if the schedule is not balanced.    */
/*************************************************************************/

float AirScheduleProblem::getLoadBalancingFactor() const 
{ return loadBalanceFactor;}

/*************************************************************************/
/** Returns the Number Of Flights to be covered.                         */
/*************************************************************************/

unsigned AirScheduleProblem::getNumberOfFlights() const 
{ return numberOfFlights;}

/*************************************************************************/
/** Returns the Number Of Cities.                                        */
/*************************************************************************/

unsigned AirScheduleProblem::getNumberOfCities() const 
{ return cIndex;}

/*************************************************************************/
/** Returns the Number Of Base Cities.                                   */
/*************************************************************************/

unsigned AirScheduleProblem::getNumberOfBaseCities() const 
{ return bIndex;}

/*************************************************************************/
/** Returns the current FlightGraph object.                              */
/*************************************************************************/

FlightGraph AirScheduleProblem::getFlightGraph() const
{ return myFlights; }

/*************************************************************************/
/** Returns the Array Of Base Cities.                                    */
/*************************************************************************/

City* AirScheduleProblem::getBaseCities() const 
{  
   // Define the return array.

   City *ret = new City[bIndex];

   // Copy the current Base City array.

   for(int i=0;i<bIndex;i++) ret[i] = BaseCities[i];

   // Return the result.

   return ret;

}// end of getBaseCities()

/*************************************************************************/
/** Returns the Overnight Stay cost.  */
/*************************************************************************/

float AirScheduleProblem::getCostPerOvernight() const 
{ return costPerOvernight;}

/*************************************************************************/
/*
// Returns the minimum amount of time we should put between consecutive
// flights.
*/
/*************************************************************************/

float AirScheduleProblem::getMinimumTime() const 
{ return minimumTime;}

/*************************************************************************/
/** Creates the FlightGraph for the current flights.  */
/*************************************************************************/

FlightGraph AirScheduleProblem::createFlightGraph() const
{
   FlightGraph ret(cIndex,cIndex);

   for(int i=0;i<numberOfFlights;i++) 
       ret.addFlight(FlightList[i]);

   return ret;

}// end of createFlightGraph()

/*************************************************************************/
/** If it is not already a base city, then adds the city as a base city. */
/*************************************************************************/

void AirScheduleProblem::addBaseCity(City _cty) 
{  
   if(!isBaseCity(_cty))
   {
      unsigned i;
      City *Temp = new City[bIndex+1];
      for(i=0;i<bIndex;i++) Temp[i] = BaseCities[i];
      City *k = new City(_cty.getName(),_cty.getIndex());
      Temp[i] = *k;
      if(bIndex) delete [] BaseCities;
      BaseCities = Temp;
      bIndex++;

   }// end of if

}// end of addBaseCity()

/*************************************************************************/
/** Checks whether the given city is Base City.  */
/*************************************************************************/

AirScheduleProblem::isBaseCity(City _cty)  const
{  unsigned i;
   int found = 0;
   for(i=0; !found && i<bIndex ;i++) if(_cty == BaseCities[i]) found=1;
   return found;
}// end of isBaseCity()

/*************************************************************************/
/**  Checks whether the given Flight starts from is a Base City.  */
/*************************************************************************/

AirScheduleProblem::isFlightFromBaseCity(Flight *F) const 
{  
   return isBaseCity(cities[F->getFromIndex()]);

}// end of isFlightFromBaseCity()

/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::isFlightToBaseCity(Flight *F) const 
{  
   return isBaseCity(cities[F->getToIndex()]);

}// end of isFlightToBaseCity()

/*************************************************************************/
/**  Checks whether the given Flight starts from is a Base City.  */
/*************************************************************************/

AirScheduleProblem::isFlightFromBaseCity(int FlightIndex) const 
{  
   int ret = 0;
   if(FlightIndex>=0 && FlightIndex<=numberOfFlights)
      ret = isBaseCity(cities[FlightList[FlightIndex].getFromIndex()]);

   return ret;

}// end of isFlightFromBaseCity()

/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::isFlightToBaseCity(int FlightIndex) const 
{  
   int ret = 0;
   if(FlightIndex>=0 && FlightIndex<=numberOfFlights)
      ret = isBaseCity(cities[FlightList[FlightIndex].getToIndex()]);

   return ret;

}// end of isFlightFromBaseCity()

/*************************************************************************/
/** Checks whether this city exists or not.  */
/*************************************************************************/

AirScheduleProblem::isCityExist(City _cty) 
{  unsigned i;
   int found = 0;
   for(i=0; !found && i<cIndex ;i++) if(_cty == cities[i]) found=1;
   return found;
}// end of isCityExist()

/*************************************************************************/
/*
//  Adds the given Flight to the FlightGraph
//  According to Departure Date.
*/
/*************************************************************************/

void AirScheduleProblem::addFlight(Flight _Flg) 
{  //myFlights.addFlight(_Flg);
   Flight *Temp = new Flight[numberOfFlights+1];
   if(numberOfFlights==0) Temp[0] = _Flg;
   else 
    {
       int i=0, j=0;
       int placed=0;
       Date  depDate = _Flg.getDepartureDate();

       while( i<numberOfFlights && FlightList[i].getDepartureDate()<depDate)
       { Temp[j] = FlightList[i]; i++; j++; }

       Temp[j]=_Flg; j++;

       while( i<numberOfFlights )
       { Temp[j] = FlightList[i]; i++; j++; }

    } 

   if(numberOfFlights) delete [] FlightList;
   //Temp[i] = _Flg;
   FlightList = Temp;
   numberOfFlights++;
   myFlights.addFlight(_Flg);
}// end of addFlight()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::updateFlightNumbers()
{
   for(int i=0;i<numberOfFlights;i++) FlightList[i].setFlightNumber(i);

}//

/*************************************************************************/
/*
// Checks this flight chain is possible. Next flight should take off 
// from the previous flight's destination.
*/
/*************************************************************************/

AirScheduleProblem::isRotationPossible(Column _col) 
{  unsigned numCities = _col.getNumberOfCities();
   unsigned index1    = _col.getIndex(1);
   unsigned index2    = _col.getIndex(numCities);
   int ret = 1;

   /***********************************************************************/
   /** Rotation should begin from the base city and end at the base city. */
   /***********************************************************************/


   ret = ret && ( isBaseCity(cities[FlightList[index1].getFromIndex()]) && 
		  isBaseCity(cities[FlightList[index1].getToIndex()]) );

   /***********************************************/
   /** Check the time between consequtive flights */
   /***********************************************/

   return ret;

}// end of isRotationPossible()

/*************************************************************************/
/*
//  Searches the given city if it finds, then it returns the index.
//  If it does not find, then it adds this city to array and returns its
//  index.
*/
/*************************************************************************/

AirScheduleProblem::lookUp(char *CityName) 
{  unsigned i=0;
   int found = 0;

   while(i<cIndex && !found )
   {  if(!strcmp(CityName,cities[i].getName())) found=1;
      else i++;
   }// end of while

   if(!found)
   {  City *Temp = new City[cIndex+1];
      City k(CityName,cIndex);
      for(i=0;i<cIndex;i++) Temp[i]=cities[i];
      if(cIndex) delete [] cities;
      Temp[i] = k;
      cities = Temp;
      i = cIndex;
      cIndex++;
   }// end of if

   return i;

}// end of lookUp()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::updateGraph()
{  myFlights.setDimensions(cIndex,cIndex);
   unsigned i=0;
   for(i=0;i<numberOfFlights;i++) myFlights.addFlight(FlightList[i]);
}// end of updateGraph()

/*************************************************************************/
/** Returns the Flight object of the given index.  */
/*************************************************************************/

Flight *  AirScheduleProblem::getFlightWithIndex(int FlightNum)
{  Flight *ret = NULL;
   if(FlightNum>=0 && FlightNum<=numberOfFlights) 
	ret = &FlightList[FlightNum];
   else 
      cout << " getFlightWithIndex() : FlightNum : " << FlightNum << " numberOfFlights : " << numberOfFlights << endl;
   return ret;
}// end of getFlightIndex()

/*************************************************************************/
/**  Returns the City object of the given index.  */
/*************************************************************************/

City *  AirScheduleProblem::getCityWithIndex(int CityNum)
{  City *ret = NULL;
   if(CityNum>=0 && CityNum<=cIndex) ret = &cities[CityNum];
   return ret;
}// end of getCityWithIndex()

/*************************************************************************/
/*************************************************************************/

AirScheduleProblem AirScheduleProblem::operator=(const AirScheduleProblem& RHS)
{  myFlights = RHS.myFlights;

   if(bIndex) delete [] BaseCities;
   bIndex = RHS.bIndex;
   BaseCities = new City[bIndex];
   for(unsigned i=0;i<bIndex;i++) BaseCities[i] = RHS.BaseCities[i];

   if(cIndex) delete [] cities;
   cIndex = RHS.cIndex;
   cities = new City[cIndex];
   for(i=0;i<cIndex;i++) cities[i] = RHS.cities[i];

      costPerMinute = RHS.costPerMinute;
   costPerOvernight = RHS.costPerOvernight;
        minimumTime = RHS.minimumTime;

   if(numberOfFlights) delete [] FlightList;

    numberOfFlights = RHS.numberOfFlights;

   FlightList = new Flight[numberOfFlights];
   for(i=0;i<numberOfFlights;i++) FlightList[i] = RHS.FlightList[i];

   return (*this);

}// end of operator=()

/*************************************************************************/
/*************************************************************************/

AirScheduleProblem::operator==(const AirScheduleProblem& RHS)
{  int ret=1;
   ret = ret && ( myFlights == RHS.myFlights);
   ret = ret && ( bIndex == RHS.bIndex);
   ret = ret && ( cIndex == RHS.cIndex);
   ret = ret && ( costPerMinute == RHS.costPerMinute );
   ret = ret && ( costPerOvernight == RHS.costPerOvernight );
   ret = ret && ( minimumTime == RHS.minimumTime );
   ret = ret && ( numberOfFlights == RHS.numberOfFlights );
   unsigned i;

   for(i=0;ret && i<bIndex ; i++) 
	ret = ret && ( BaseCities[i] == RHS.BaseCities[i] );

   for(i=0;ret && i<cIndex ; i++) 
	ret = ret && ( cities[i] == RHS.cities[i] );

   for(i=0;ret && i<numberOfFlights ; i++) 
	ret = ret && ( FlightList[i] == RHS.FlightList[i] );

   return ret;

}// end of operator==()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::findBaseCities() 
{
   int i, from, to, stay, f1, f2;
   myQueue<int> *inComing = new myQueue<int>[cIndex];
   myQueue<int> *outGoing = new myQueue<int>[cIndex];

   // For each flight
   for(i=0;i<numberOfFlights;i++)
   {
      from = FlightList[i].getFromIndex();
      to   = FlightList[i].getToIndex();

      inComing[to].addItem(i);
      outGoing[from].addItem(i);

      stay = 1;

      while( stay && inComing[from].getSize()>0 && outGoing[from].getSize()>0 )
	if( inComing[from].peekHead(f1) && outGoing[from].peekHead(f2) &&
	    FlightList[f1].getArrivalDate() < FlightList[f2].getDepartureDate() )
        {
	    inComing[from].removeHead(f1);
	    outGoing[from].removeHead(f1);
	}
	else stay = 0;

      stay = 1;

      while( stay && inComing[to].getSize()>0 && outGoing[to].getSize()>0 )
	if( inComing[to].peekHead(f1) && outGoing[to].peekHead(f2) &&
	    FlightList[f1].getArrivalDate() < FlightList[f2].getDepartureDate() )
        {
	    inComing[to].removeHead(f1);
	    outGoing[to].removeHead(f1);
	}
	else stay = 0;
      
   }// end of for

   for(i=0;i<cIndex;i++)
     if( inComing[i].getSize() == outGoing[i].getSize() &&
	 inComing[i].getSize()>0 ) addBaseCity(cities[i]);


   delete[] inComing;
   delete[] outGoing;

}// end of findBaseCities()

/*************************************************************************/
/*************************************************************************/

ostream & operator<<(ostream&os, const AirScheduleProblem& RHS)
{  os << " Number Of Flights : " << RHS.numberOfFlights << endl;
   os << " Cost Per Minute   : " << RHS.costPerMinute   << endl;
   os << " Cost Per Overnight : " << RHS.costPerOvernight << endl;
   os << " Miniumum Wait Time : " << RHS.minimumTime << endl;
   os << " Base Cities : ";
   for(unsigned i=0;i<RHS.bIndex;i++) os << RHS.BaseCities[i];
   os << endl;
   os << " Cities : ";
   for(i=0;i<RHS.cIndex;i++) os << RHS.cities[i];
   os << endl;
   os << " Flights : " << endl;
   for(i=0;i<RHS.numberOfFlights;i++) 
       os << setw(3) << i << " " << RHS.FlightList[i] << endl;
   os << RHS.myFlights << endl;

   os << " Total Flying time : " << RHS.getTotalFlyingTime() << endl;

   return os;

}// end of operator<<()

/*************************************************************************/
/**
*  Finds the minimum number of rotations we need to cover the flight
*  schedule. Note that this gives only a theoretical lower bound.
*  Regulations won't allow this number (I think). But it will give us
*  some comparison anchor.
*/
/*************************************************************************/
AirScheduleProblem::getLowerBound() 
{  
   Flight *f;
   int max = 1, curr = 0, i, j, k;
   Date *sorted = new Date[2*numberOfFlights];
   int  *sortIndex = new int[2*numberOfFlights];

   k=0;
   for(i=0;i<numberOfFlights;i++)
   {
      f = getFlightWithIndex(i);

      if(f!=NULL)
      {
	  sorted[2*k] = f->getDepartureDate();
	  sorted[2*k+1] = f->getArrivalDate();
	  sortIndex[2*k] = k;
	  sortIndex[2*k+1] = k;
	  k++;

      }// end of if

   }// end of for

   /********************************************************/
   /** We hve k number of flights                          */
   /** Lets' sort them according to arrival/departure time */
   /** O(k^2)                                              */
   /********************************************************/
   Date dummy;
   int  intDummy;
   for(i=0;i<2*k;i++)
       for(j=i+1;j<2*k;j++)
	  if(sorted[i]>sorted[j])
          {
	     cout << " replace " << endl;
	     //swap(sorted[i],sorted[j]);
	     //swap(sortIndex[i],sortIndex[j]);
	     dummy = sorted[j];
	     sorted[j] = sorted[i];
	     sorted[i] = dummy;

	     intDummy = sortIndex[j];
	     sortIndex[j] = sortIndex[i];
	     sortIndex[i] = intDummy;

          }// end of if

   for(i=0;i<2*k;i++) cout << sorted[i] << " " << sortIndex[i] << endl;

   /***********************************/
   /** We need to examine each flight */
   /***********************************/
   char *covered = new char[k];
   for(i=0;i<k;i++) covered[i]=0;

   i = curr = 0 ;

   /*****************************************************/
   /** We need to go through all the flights - O(k*max) */
   /*****************************************************/
   while( i < (2*k) )
   {
      /****************************/
      /** number of crews we need */
      /****************************/
      curr = 1;
      j = i + 1;

      /************************************************************/
      /** If we did not check this flight before                  */
      /** If we did, then this value represents the arrival time, */
      /** and we do not care about this time                      */
      /************************************************************/
      if(covered[sortIndex[i]]==0)
      {
         while( j<(2*k) && sortIndex[j]!=sortIndex[i] ) 
	 { 
	    j++; 
	    f = getFlightWithIndex(sortIndex[j]);

	    if(sorted[j]>f->getDepartureDate()) 
	    {
               max = maximum(max,curr);
	       curr--;

	    } else curr++; 

	 }//

	 covered[sortIndex[i]]=1;

      }// end of if

      /*******************************************/
      /** Compare with the current maximum value */
      /*******************************************/
      max = maximum(max,curr);
      i++;

   }// end of while

   return max;

}// end of getLowerBound()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::resetDifficulties()
{
   for(int i=0;i<numberOfFlights;i++) difficulty[i] = 1.0;

}// end of resetDifficulties()

/*************************************************************************/
/*************************************************************************/

float AirScheduleProblem::getDifficulty(int indx)
{
  int ret = 0.0;
  if(indx>=0 && indx<numberOfFlights) ret = difficulty[indx];
  return ret;
}// end of getDifficulty()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::incrementDifficulty(int indx)
{
   if(indx>=0 && indx<numberOfFlights) difficulty[indx]++;

}// end of incrementDifficulty()

/*************************************************************************/
/*************************************************************************/

void AirScheduleProblem::buildPreSolutions()
{
   int i;
   PathBuilder pbuilder(this);
   sol = new ListOfEdgeLists[numberOfFlights];

   for(i=0;i<numberOfFlights;i++) 
     pbuilder.buildCyclesN(*myEdgeGraph,sol[i],i,getMaxRotationPeriod(),3,1);

   difficulty = new float[numberOfFlights];
   for(i=0;i<numberOfFlights;i++) difficulty[i] = 1.0;

}// end of buildPreSolutions()

/*************************************************************************/
/**
    Updates the solution based on minium per minute cost.
**/
/*************************************************************************/

/*
void AirScheduleProblem::updatePreSolutions(EdgeList* eList, int fi)
{

   EdgeList *eList1 = NULL;
   int j,k,selected = -1;
   float ccost=0, cost=0, minutes=0;
   int stay=1;

   if(fi>=0 && fi<numberOfFlights)
   {
      cost = eList->getDCost(this)/eList->getTotalFlyingTime();

      for(j=0;stay&&j<sol[fi].getNumberOfEdgeLists();j++)
      {
	 eList1 = sol[fi].getNthEdgeList(j);

	 if( eList1!=NULL )
	 {
            ccost = eList1->getDCost(this)/eList1->getTotalFlyingTime();

	    if( ccost > cost ) { stay=0; *eList1=*eList; }

	 }// end of if

      }// end of for

   }// end of if

}// end of updatePreSolutions()
*/

void AirScheduleProblem::updatePreSolutions( EdgeListPtr eList )
{

   EdgePtr  edge;
   EdgeList *eList1 = NULL;
   int i, j, k, fn, selected = -1;
   float ccost=0, cost=0 ;
   int stay=1;

   cost = eList->getDCost(this)/eList->getTotalFlyingTime();

   for(i=0;i<eList->getSize();i++)
   {
     edge = eList->getNthEdge(i);
     if( edge!=NULL )
     {
        fn = edge->getStartNode();

        stay=1;
        for(j=0;stay&&j<sol[fn].getNumberOfEdgeLists();j++)
        {

          eList1 = sol[fn].getNthEdgeList(j);

          if( eList1!=NULL && eList->getNumberOfDifferences(eList1)>0 )
          {
               ccost = eList1->getDCost(this)/eList1->getTotalFlyingTime();

	       if( ccost > cost ) { stay=0; *eList1=*eList; }

           }// end of if

         }// end of for

     }// end of if

   }// end of for

}// end of updatePreSolutions()

/*************************************************************************/
/**
    Assume that we have N solutions in the current solution list. Now,
    the question is how to selec the best suitable one based on the
    current uncovered nodes and the cost:

      cost = total_cost/(new_covered_minutes)
  
    This formulae gives the per minute cost of covering uncovered
    flights. Therefore, a solution which only covers one node with cost
    A, compared with a solution that covers more than one node with cost
    B, where B>A. Simple cost based selection decides in favour of the
    first solution. But, there might be the case that B the second
    solution covers more with less cost even B>A.

**/
/*************************************************************************/

EdgeList *AirScheduleProblem::selectForNode(int fi, char *covered)
{

   EdgeList *ret = NULL;
   EdgeList *eList = NULL;
   int j,k,selected = -1;
   float ccost=0, cost=0, minutes=0;

   if(fi>=0 && fi<numberOfFlights)
   {
      for(j=0;j<sol[fi].getNumberOfEdgeLists();j++)
      {
	 eList = sol[fi].getNthEdgeList(j);

	 /*minutes = 0;
	  
	 for(k=0;k<numberOfFlights;k++)
	   if( covered[k]==0 && eList->doesContainNodeId(k) ) 
	       minutes += FlightList[k].getLength();
         ccost = eList->getDCost(this)/minutes;
	 */

         ccost = eList->getCostPerUncoveredMinute(this,covered);

	 if( selected==-1 || ccost < cost )
	 { selected = j; cost=ccost; }

      }// end of for

      if( selected!=-1 )
      {
	 ret = sol[fi].getNthEdgeList(selected);
	 EdgeList *a = new EdgeList(ret->getNumberOfEdges());
	 *a = *ret;
	 ret=a;

      }// end of if

   }// end of if

   return ret;

}// end of selectForNode()

